using static System.Console;
using static System.Math;

using Cairo;
using Gdk;
using Gtk;
using CairoHelper = Gdk.CairoHelper;
using Timeout = GLib.Timeout;

static class Util {
    public static RGBA parse(string name) {
        RGBA c = new();
        if (!c.Parse(name))
            throw new Exception("bad color");
        return c;
    }
    
    public static void setColor(this Context c, RGBA color) {
        CairoHelper.SetSourceRgba(c, color);
    }
}

class GameArea : DrawingArea {
    const int SQUARE_SIZE = 100;
    const int MARGIN = SQUARE_SIZE / 2;
    
    static RGBA[] colors = {
        Util.parse("white"),
        Util.parse("#D0D016"),  // yellow
        Util.parse("#AA190B")   // red 
    };
    
    Game game = new(false);
    Player ai;
    int last_x = -1, last_y = -1;
    double drop_y = -1.0;
    
    void init() {
        game = new Game(false);
        QueueDraw();     // draw the initial position
    }
    
    public GameArea(Player ai) {
        this.ai = ai;
        
        init();
        CanFocus = true;
        AddEvents((int) EventMask.ButtonPressMask);
    }

    void move(int x) {
        int y = game.move(x);
        if (y >= 0) {
            (last_x, last_y) = (x, y);
            drop_y = 0;
            Timeout.Add(20, tick);
            QueueDraw();
        }
    }

    bool tick() {
        drop_y += 0.15;
        if (drop_y >= last_y + 0.5) {
            drop_y = -1.0;
            if (game._winner == null && game.turn == 2) {
                move(ai.move(game));
            }
        }
        QueueDraw();

        return drop_y >= 0;
    }
    
    protected override bool OnButtonPressEvent(EventButton ev) {
        if (ev.Type == EventType.ButtonPress)
            if (game._winner != null) {
                init();     // start a new game
            }
            else {
                int x = ((int) (ev.X - MARGIN)) / SQUARE_SIZE;
                move(x);
            }

        return true;
    }

    protected override bool OnDrawn(Context c) {
        // Set a transformation matrix so that squares are 1 unit tall/wide, with a
        // MARGIN around the drawing area.
        c.Translate(MARGIN, MARGIN);
        c.Scale(SQUARE_SIZE, SQUARE_SIZE);
        
        // Adjust the line width to compensate for the scaling tranformation.
        c.LineWidth /= SQUARE_SIZE;
        
        for (int x = 0 ; x < Game.Width ; ++x)
            for (int y = 0 ; y < Game.Height ; ++y) {
                double yp = y + 0.5;
                if (x == last_x && y == last_y && drop_y >= 0) {
                    if (drop_y >= y + 0.5)
                        drop_y = -1.0;
                    else
                        yp = drop_y;
                }
                c.setColor(colors[game[x, y]]);
                c.Arc(x + 0.5, yp, 0.4, 0, 2 * PI);
                c.Fill();
            }

        // draw frame, without holes
        c.setColor(Util.parse("dark blue"));
        c.Rectangle(0, 0, Game.Width, Game.Height);
        for (int x = 0 ; x < Game.Width ; ++x)
            for (int y = 0 ; y < Game.Height ; ++y) {
                c.NewSubPath();
                c.ArcNegative(x + 0.5, y + 0.5, 0.4, 0, - 2 * PI);
            }
        c.Fill();
        
        if (game._winner > 0 && drop_y < 0) {
            for (int i = 0 ; i < 4 ; ++i) {
                c.setColor(Util.parse("light green"));
                int x = game.win_x + i * game.win_dx, y = game.win_y + i * game.win_dy;
                c.Rectangle(x,y, 1, 1);
                c.Fill();
                c.setColor(colors[game[x, y]]);
                c.Arc(x + 0.5, y + 0.5, 0.4, 0, 2 * PI);
                c.Fill();
            }
        }
        
        return true;
    }

}

public class MainWindow : Gtk.Window {
    public MainWindow(Player ai)
        : base($"Connect Four: yellow = human | red = {ai}") {
        Resize(800, 700);
        GameArea area = new(ai);
        Add(area);
        area.GrabFocus();
    }
    
    protected override bool OnDeleteEvent(Event ev) {
        Application.Quit();
        return true;
    }
}

class ConnectFour {
    static Player player(string name) {
        if (name == "random")
            return new RandomPlayer();

        string prefix = "minimax:";
        if (name.StartsWith(prefix)) {
            int max_depth = int.Parse(name[prefix.Length..]);
            return new MinimaxPlayer(max_depth);
        }

        throw new Exception("unknown player");
    }

    static void tournament(int games, string name1, string name2) {
        string[] names = { "draw", name1, name2 };
        Player?[] players = { null, player(name1), player(name2) };
        int[] won = new int[3];

        for (int i = 0 ; i < games ; ++i) {
            Game g = new(true);
            while (true) {
                if (g._winner is int w) {
                    won[w] += 1;
                    WriteLine($"game {i}: winner = {names[w]}");
                    break;
                }
                g.move(players[g.turn]!.move(g));
            }
        }
        WriteLine($"{names[1]} won {won[1]}, {names[2]} won {won[2]}, {won[0]} draws");
    }

    static void Main(string[] args) {
        if (args.Length == 0) {
            WriteLine("usage: connect_four <ai_player> | -g <count> <ai_1> <ai_2>");
            return;
        }
        if (args.Length > 0 && args[0] == "-g") {
            int games = int.Parse(args[1]);
            tournament(games, args[2], args[3]);
        } else {
            Player ai = player(args[0]);
            Application.Init();
            new MainWindow(ai).ShowAll();
            Application.Run();
        }
    }
}
